#include "fishBMC.h"

#include <cstdlib>
#include <unistd.h>
#include <ctime>

#include "vectorfield.h"
#include "configuration.h"
#include "screenbuffer.h"
#include "busyindicator.h"
#include "blurengine.h"
#include "wavepainter.h"
#include "analyst.h"
#include "ringbuffer.h"

#include "xbmc/addons/include/xbmc_addon_dll.h"
#include "xbmc/addons/include/xbmc_vis_types.h"
#include "xbmc/utils/TimeUtils.h"

FishBMC::FishBMC (VIS_PROPS* properties)
{
	// (re-)init random number generator
	srand (time (NULL));

	// create an empty configuration
	mp_config			= Configuration::instance();

	// get aspect ratio
	double aspect		= (double) properties->width / (double) properties->height;
	mp_config->setAspectRatio (aspect);

	// texture will have an aspect of 2:1, calculate tex coordinates to correct for that
	m_TextureLeft		= (2 - aspect) / 4;
	m_TextureRight		= 1 - m_TextureLeft;

	DEBUGOUT ("new fishBMC");
}

FishBMC::~FishBMC()
{
	mp_config->release();
}

void FishBMC::stop(){
	m_CancelInit = true;
	if (mp_VectorThread) SDL_WaitThread (mp_VectorThread, NULL);
	mp_VectorThread = NULL;

	m_VectorProgress = 1;
	if (mp_BusyThread) SDL_WaitThread (mp_BusyThread, NULL);
	mp_BusyThread = NULL;

	glDeleteTextures (1, &m_TextureID);

	mp_ringbuffer->release();
	mp_analyst->release();
	mp_wavepainter->release();
	mp_blurengine->release();
	mp_vectorfield->release();
	mp_screenbuffer->release();
}

void FishBMC::start()
{
	// create a pixel buffer (32bit little endian RGBA)
	mp_screenbuffer		= ScreenBuffer::instance (mp_config->width(), mp_config->height());
	void* pixels		= mp_screenbuffer->pixels();

	// generate a texture for drawing into
	glEnable (GL_TEXTURE_2D);
	glGenTextures (1, &m_TextureID);
	glBindTexture (GL_TEXTURE_2D, m_TextureID);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, mp_config->width(), mp_config->height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, pixels);


	// indicate initialisation busy
	m_VectorProgress	= 0;
	mp_BusyThread		= SDL_CreateThread (busy_indicator, &m_VectorProgress);
	m_InitBusy			= true;

	m_CancelInit		= false;

	// start creating vectors
	mp_vectorfield = NULL;
	mp_VectorThread		= SDL_CreateThread (FishBMC::makevectors, this);

	// initialize remaining components
	mp_blurengine		= BlurEngine::instance();
	mp_wavepainter		= WavePainter::instance();
	mp_analyst			= Analyst::instance();

	// let the ringbuffer hold the last tenth of a second
	mp_ringbuffer		= RingBuffer::instance (17640);

	// initialize frame counter
	m_FrameCounter		= 0;
}

int FishBMC::makevectors (void* p)
{
	FishBMC* f = (FishBMC*) p;
	VectorField::instance (f->mp_config->width(), f->mp_config->height(), & (f->m_VectorProgress), & (f->m_CancelInit));
	return 0;
}

void FishBMC::paint_frame()
{
	// determine time since last frame
	static int last_frame_time	= CTimeUtils::GetTimeMS();
	double time_passed			= (double) (CTimeUtils::GetTimeMS() - last_frame_time) / 1000;
	last_frame_time				= CTimeUtils::GetTimeMS();
	
	if (!m_InitBusy) {
		// get audio data from ringbuffer
		int audiosize = 0;
		int16_t* audiodata = (int16_t*) mp_ringbuffer->get (time_passed, &audiosize);

		// analyse audio
		int analysis = mp_analyst->analyse (audiodata, audiosize / 2);

		// act accordingly
		if (mp_config->nervous()) {
			if (analysis >= 2) mp_wavepainter->changeShape();
			if (analysis >= 1) mp_vectorfield->change();
		} else {
			if (analysis >= 1) mp_wavepainter->changeShape();
			if (analysis >= 2) mp_vectorfield->change();
		}

		if (analysis >= 3) mp_wavepainter->beat (mp_analyst->framesPerBeat());

		mp_wavepainter->changeColor (mp_analyst->framesPerBeat(), mp_analyst->relativeEnergy());
		if (analysis != -1)
			mp_wavepainter->paint (audiodata, audiosize / 4);
	}
}

void FishBMC::render()
{
	try_init_complete();

	// check if this frame is to be skipped
	if (m_FrameCounter % mp_config->frameDivisor() == 0) {
		paint_frame();
	}

	// actual rendering
	renderGL();

	// increment frame counter an do the heavy lifting (=blur)
	if ( (++m_FrameCounter % mp_config->frameDivisor() == 0) && (!m_InitBusy))
		mp_blurengine->blur();
}

void FishBMC::renderGL()
{
	// OpenGL settings
	glEnable (GL_BLEND);
	glEnable (GL_TEXTURE_2D);
	glDisable (GL_DEPTH_TEST);
	glPolygonMode (GL_FRONT, GL_FILL);

	// OpenGL matrix setup
	glMatrixMode (GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho (-1, 1, -1, 1, -1, 1);

	glMatrixMode (GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	// actual painting
	glBindTexture (GL_TEXTURE_2D, m_TextureID);

	// get a lock on the texture pixels
	while (!mp_screenbuffer->lock())
		usleep (1);

	// update texture
	void* pixels = mp_screenbuffer->pixels();
	glTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, mp_config->width(), mp_config->height(), GL_RGBA, GL_UNSIGNED_BYTE, pixels);
	
	// release the pixel buffer
	mp_screenbuffer->unlock();

	// a textured rectangle spanning the whole window
	glBegin (GL_QUADS);
	glTexCoord2f (m_TextureLeft, 0.0f);
	glVertex3f (-1.0f, -1.0f, 1.0f);
	glTexCoord2f (m_TextureRight, 0.0f);
	glVertex3f (1.0f, -1.0f, 1.0f);
	glTexCoord2f (m_TextureRight, 1.0f);
	glVertex3f (1.0f, 1.0f, 1.0f);
	glTexCoord2f (m_TextureLeft, 1.0f);
	glVertex3f (-1.0f, 1.0f, 1.0f);
	glEnd();

	// OpenGL matrix to original state
	glPopMatrix();
	glMatrixMode (GL_PROJECTION);
	glPopMatrix();
}

void FishBMC::try_init_complete()
{
	if (m_InitBusy) {
		if (m_VectorProgress > 0.97) {
			m_VectorProgress = 1;
			SDL_WaitThread (mp_BusyThread, NULL);
			mp_BusyThread = NULL;

			SDL_WaitThread (mp_VectorThread, NULL);
			mp_VectorThread = NULL;
			mp_vectorfield = VectorField::instance();

			m_InitBusy = false;
		}
	}
}

void FishBMC::newData (const void* data, int size)
{
	mp_ringbuffer->insert ( (char*) data, size);
}
