
/*
 * fische-3.2
 * Copyright (C) Marcel Ebmer 2009-2010 <marcel@26elf.at>
 *
 * fische-3.2 is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * fische-3.2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "fishBMC.h"

#include <cstdlib>
#include <cstdio>
#include <unistd.h>
#include <ctime>
#ifdef WIN32
#include <SDL.h>
#else
#include <SDL/SDL.h>
#endif
#include "vectorfield.h"
#include "configuration.h"

#ifdef WIN32
#define rand_r( _seed ) ( _seed == _seed? rand() : rand() )
#endif

#define N_FIELDS 20

using namespace std;

VectorField* VectorField::ms_instance = 0;
const double Pi = acos (-1);
unsigned int rand_seed = time (NULL);

VectorField::VectorField (int x, int y, double* progress, bool* cancel)
{
	m_Width = x;
	m_Height = y;
	m_Diagonal = (int) (.5 + sqrt (pow ( (double) m_Width, 2) + pow ( (double) m_Height, 2)));
	m_CenterX = m_Width / 2;
	m_CenterY = m_Height / 2;

	mp_ExportData = NULL;

	if (progress) *progress = 0;

	int n = 0;
	uint16_t* new_field = fillField (n);
	if (progress) *progress = 1.0f / N_FIELDS;

	bool* c = cancel;
	if (!cancel)
		c = new bool (false);

	while (new_field && !*c) {
		m_Fields.push_back (new_field);
		n++;
		new_field = fillField (n);

		if (progress) *progress = ( (double) (n + 1)) / N_FIELDS;
	}

	if (!cancel)
		delete c;

	mp_ExportData = m_Fields.back();
	DEBUGOUT ("new vectorfield");
}

VectorField::~VectorField()
{
	for (unsigned int i = 0; i < m_Fields.size(); i++) {
		delete[] m_Fields.at (i);
	}
	m_Fields.clear();
	DEBUGOUT ("delete vectorfield");
}

VectorField* VectorField::instance (int x, int y, double* progress, bool* cancel)
{
	if (!ms_instance) ms_instance = new VectorField (x, y, progress, cancel);
	return ms_instance;
}

void VectorField::release()
{
	delete ms_instance;
	ms_instance = 0;
}

char* VectorField::get()
{
	return (char*) mp_ExportData;
}

void VectorField::change()
{
	int n = rand_r (&rand_seed) % m_Fields.size();
	while (mp_ExportData == m_Fields[n]) n = rand_r (&rand_seed) % m_Fields.size();
	mp_ExportData = m_Fields[n];
}

int VectorField::width() const
{
	return m_Width;
}

int VectorField::height() const
{
	return m_Height;
}

void VectorField::randomize (Vector& vec, int n)
{
	return;
	if (n % 2 == 0) n++;
	Vector randv ( (rand_r (&rand_seed) % n) - n / 2, (rand_r (&rand_seed) % n) - n / 2);
	vec += randv;
}

struct field_param {
	uint16_t* field;
	int number;
	int start_y;
	int end_y;
	VectorField* vecfield;
};

int VectorField::fill_thread (void* param)
{
	struct field_param* p = (struct field_param*) param;
	uint16_t* field = p->field;
	int f = p->number;
	VectorField* vf = p->vecfield;

	for (int y = p->start_y; y < p->end_y; y ++) {
		for (int x = 0; x < vf->m_Width; x++) {
			uint16_t* vector = field + x + y * vf->m_Width;

			Vector p (x - vf->m_CenterX, y - vf->m_CenterY);
			Vector e = p.singleUnit();
			Vector n = e.normal();
			double r = p.length();

			Vector p1;
			if (vf->m_Width >= vf->m_Height) p1 = Vector (x - vf->m_CenterX / 3, y - vf->m_CenterY);
			else p1 = Vector (x - vf->m_CenterX, y - vf->m_CenterY / 3);
			Vector e1 = p1.singleUnit();
			Vector n1 = e1.normal();
			double r1 = p1.length();

			Vector p2;
			if (vf->m_Width >= vf->m_Height) p2 = Vector (x - vf->m_Width * 5 / 6, y - vf->m_CenterY);
			else p2 = Vector (x - vf->m_CenterX, y - vf->m_Height * 5 / 6);
			Vector e2 = p2.singleUnit();
			Vector n2 = e2.normal();
			double r2 = p2.length();

			Vector v;
			switch (f) {
				case 0:
					// linear vectors showing away from a horizontal mirror axis
					v = Vector (0, (y < vf->m_CenterY) ? (vf->yMax() / 100) : (-vf->yMax() / 100));
					break;
				case 1:
					// linear vectors showing away from a vertical mirror axis
					v = Vector ( (x < vf->m_CenterX) ? (vf->xMax() / 100) : (-vf->xMax() / 100), 0);
					break;
				case 2:
					// radial vectors showing away from the center
					v = e * -r / 25;
					break;
				case 3:
					// tangential vectors (right)
					v = n * r / 25;
					break;
				case 4:
					// tangential vectors (left)
					v = n * -r / 25;
					break;
				case 5:
					// tangential-radial vectors (left)
					v = e * -r / 25 - n * r / 25;
					break;
				case 6:
					// tangential-radial vectors (right)
					v = e * -r / 25 + n * r / 25;
					break;
				case 7: {
					// tree rings
					double dv = cos (Pi * 20 * r / vf->dimension());
					v = e * 5 * dv;
					break;
				}
				case 8: {
					// hyperbolic vectors
					Vector _e (e.y(), e.x());
					v = _e * -r / 25;
					break;
				}
				case 9: {
					// hyperbolic vectors
					Vector _e (e.y(), e.x());
					v = _e * r / 25;
					break;
				}
				case 10:
					// purely random
					v = Vector (rand_r (&rand_seed) % 21 - 10, rand_r (&rand_seed) % 21 - 10);
					break;
				case 11: {
					// sphere
					double dv = cos (Pi * r / vf->dimension());
					v = e * -r * dv / 15;
					break;
				}
				case 12: {
					// sine distortion
					double dv = sin (Pi * 6 * r / vf->dimension());
					v = e * -r * dv / 25;
					break;
				}
				case 13:
					// black hole
					v = n * vf->dimension() / 50 + e * r / 50;
					if (r < 10) v = Vector (0, 0);
					break;
				case 14:
					// black hole
					v = n * -vf->dimension() / 50 + e * r / 50;
					if (r < 10) v = Vector (0, 0);
					break;
				case 15: {
					// circular waves
					double dim = 121 * pow (Pi, 2) / vf->dimension();
					double vx = r * dim;
					v = e * -10 * sqrt (1.04 - pow (cos (sqrt (vx)), 2) + 0.25 * pow (sin (sqrt (vx)), 2));
					break;
				}
				case 16:
					// spinning CD
					if (fabs (r - (vf->dimension() / 4)) < vf->dimension() * 1 / 6) v = n * r / 25;
					//if ( (int) (sqrt (sqrt (r * 200 / dimension() ) ) ) % 2 == 0) v = n * r / 25;
					else v = n * -r / 25;
					break;
				case 17:
					// three spinning disks
				{
					double rt = (vf->m_Width > vf->m_Height) ? vf->m_Width / 6 : vf->m_Height / 6;
					if (r < rt * 1.2) {
						v = n * -r / 15;
					} else if (r1 < rt) {
						v = n1 * r1 / 15;
					} else if (r2 < rt) {
						v = n2 * r2 / 15;
					} else v = Vector (rand_r (&rand_seed) % 21 - 10, rand_r (&rand_seed) % 21 - 10);
					break;
				}
				case 18:
					// experimental
					v = e1 * (vf->diagonal() - r1) / 100 + e2 * (vf->diagonal() - r2) / 100 + e * (vf->diagonal() - r) / -100;
					break;
				case 19:
					// experimental
					v = n1 * (vf->diagonal() - r1) / -100 + n2 * (vf->diagonal() - r2) / -100 + n * (vf->diagonal() - r) / 100;
					break;
				default:
					// index too high. return nothing.
					return 1;
			}
			vf->randomize (v, 2);
			vf->validate (v, x, y);
			*vector = v.toUint16();
		}
	}
	return 0;
}

uint16_t* VectorField::fillField (int f)
{
	uint16_t* field = new uint16_t[m_Width * m_Height];

	if (!field) return NULL;

	int n = Configuration::instance()->processorCores();
	if (n > 8) n = 8;
	if (n % 2 != 0) n--;
	if (n == 0) n = 1;

	SDL_Thread* t[n];

	struct field_param p[n];
	for (int i = 0; i < n; i ++) {
		p[i].field = field;
		p[i].number = f;
		p[i].start_y = (i * m_Height) / n;
		p[i].end_y = ( (i + 1) * m_Height) / n;
		p[i].vecfield = this;

		t[i] = SDL_CreateThread (fill_thread, &p[i]);
	}
	int s;
	for (int i = 0; i < n; i ++) {
		SDL_WaitThread (t[i], &s);
	}
	if (s) {
		delete[] field;
		return NULL;
	}

	return field;
}

int VectorField::dimension() const
{
	return (m_Width > m_Height) ? m_Height : m_Width;
}

int VectorField::diagonal() const
{
	return m_Diagonal;
}
